from .workflow import SoftwareCompanyWorkflow
from .main import run_workflow
from .agent_adapter import CLIAgent, create_gemini_agent, create_amazonq_agent, create_opencode_agent
from .cli.base import BaseCLIWrapper
from .cli.gemini import GeminiCLIWrapper
from .cli.amazonq import AmazonQCLIWrapper
from .cli.opencode import OpenCodeCLIWrapper

__all__ = [
    "SoftwareCompanyWorkflow",
    "run_workflow",
    "CLIAgent",
    "create_gemini_agent",
    "create_amazonq_agent",
    "create_opencode_agent",
    "BaseCLIWrapper",
    "GeminiCLIWrapper",
    "AmazonQCLIWrapper",
    "OpenCodeCLIWrapper"
]


import os
from .state_manager import StateManager

def generate_timeline_report(task_id: str, state_manager: StateManager):
    """
    Generates a Markdown report of the task execution timeline.
    """
    timeline = state_manager.get_timeline()
    if not timeline:
        # We might want to generate an empty report or just skip.
        # But if the task completed, there should be a timeline.
        # If it crashed immediately, maybe not.
        return

    # Table Header
    content = f"# Workflow Execution Timeline - Task {task_id}\n\n"
    content += "| Sequence ID | Step Name | Role/Tool | Status | Start Time | End Time | Duration |\n"
    content += "|---|---|---|---|---|---|---|\n"

    for entry in timeline:
        seq_id = entry.get("sequence_id", "")
        step_name = entry.get("step_name", "")
        role_tool = entry.get("role_tool", "")
        status = entry.get("status", "")
        start_time = entry.get("start_time", "")
        end_time = entry.get("end_time", "")
        duration = entry.get("duration", "")

        content += f"| {seq_id} | {step_name} | {role_tool} | {status} | {start_time} | {end_time} | {duration} |\n"

    # Save to file
    report_filename = "timeline.md"
    report_path = os.path.join(state_manager.task_dir, report_filename)

    try:
        with open(report_path, "w", encoding='utf-8') as f:
            f.write(content)
        # We can assume UI is available or print to stdout.
        # Since this is a library function, maybe it shouldn't print,
        # but the requirements say "Modify... to ensure... report generation is triggered".
        # WorkflowEngine has UI, so maybe we can let WorkflowEngine log the success message.
    except Exception as e:
        print(f"Error generating timeline report: {e}")



import asyncio
import argparse
import os
import hashlib
import sys
from datetime import datetime
from .workflow import SoftwareCompanyWorkflow, WorkflowStep, ALL_POSSIBLE_STEPS, TOOL_FACTORIES
from .config import load_config
from .utils import list_tools, list_tasks
from .rich_help import RichArgumentParser

async def run_workflow():
    parser = RichArgumentParser(description="Software Company Team Workflow")
    parser.add_argument("prompt", nargs="?", help="The feature request or task description")
    parser.add_argument("--id", help="Unique ID for the task (resumes if exists)", default=None)
    parser.add_argument("--config-file", help="Path to a custom configuration file", default=None)
    parser.add_argument("--template", help="Name of a workflow template to use (e.g., planning, development)", default=None)
    parser.add_argument("--revert", action="store_true", help="Revert the last completed step of the specified task")
    parser.add_argument("--step", action="store_true", help="Execute only the next single step and then stop")
    parser.add_argument("--run-step", help="Jump to and run a specific step by name")
    parser.add_argument("--list-tools", action="store_true", help="List available AI tools")
    parser.add_argument("--list-tasks", action="store_true", help="List previous tasks and their status")
    parser.add_argument("--tui", action="store_true", help="Run in Terminal User Interface mode (experimental)")

    # We use default config values just for parser help/defaults logic,
    # but actual values will come from loaded config.
    default_tool_name = "gemini"

    parser.add_argument("--default-tool", help="Default tool for all roles", default=None, choices=TOOL_FACTORIES.keys())
    parser.add_argument("--default-model", help="Default model for all roles", default=None)
    parser.add_argument("--work-dir", help="Working directory for the agents", default=None)

    # Add arguments for each step dynamically from the FULL set of possible steps (from package default)
    for step in ALL_POSSIBLE_STEPS:
        parser.add_argument(f"--{step.arg_name}", help=f"Tool for {step.name}", choices=TOOL_FACTORIES.keys())
        parser.add_argument(f"--{step.model_arg_name}", help=f"Model for {step.name}")
        parser.add_argument(f"--{step.fallback_tool_arg_name}", help=f"Fallback tool for {step.name}", choices=TOOL_FACTORIES.keys())
        parser.add_argument(f"--{step.fallback_model_arg_name}", help=f"Fallback model for {step.name}")

    args = parser.parse_args()

    config_path = args.config_file

    if args.template:
        template_name = args.template
        if not template_name.endswith(".json"):
            template_name += ".json"

        base_dir = os.path.dirname(os.path.abspath(__file__))
        template_dir = os.path.join(base_dir, "templates")
        potential_path = os.path.join(template_dir, template_name)

        if os.path.exists(potential_path):
            if config_path:
                print(f"Warning: --template '{args.template}' is overriding --config-file '{args.config_file}'.")
            config_path = potential_path
            print(f"Using template: {template_name}")
        else:
            print(f"Error: Template '{args.template}' not found.")
            print("Available templates:")
            if os.path.exists(template_dir):
                for f in sorted(os.listdir(template_dir)):
                    if f.endswith(".json"):
                         print(f"  - {f[:-5]}")
            else:
                print("  (Templates directory not found)")
            sys.exit(1)

    # Load Config (Priority: CLI --config-file > CWD config.json > Default)
    config_data = load_config(config_path)

    # Runtime Steps
    runtime_steps = [WorkflowStep.from_dict(step_data) for step_data in config_data.get("steps", [])]
    state_directory = config_data.get("state_directory", ".tasks") # Default to .tasks if not in config

    if args.list_tools:
        list_tools(list(TOOL_FACTORIES.keys()))
        return

    if args.list_tasks:
        list_tasks(state_directory)
        return

    max_workflow_iterations = config_data.get("max_workflow_iterations", 20)
    default_timeout = config_data.get("default_timeout", 120)

    # Resolve Defaults (CLI Arg > Config > Hardcoded 'gemini')
    default_tool = args.default_tool or config_data.get("default_tool", "gemini")
    default_model = args.default_model or config_data.get("default_model", None)
    work_dir = args.work_dir or config_data.get("default_work_dir", None)

    task_id = args.id
    prompt = args.prompt

    if work_dir:
        work_dir = os.path.abspath(work_dir)
        if not os.path.exists(work_dir):
            try:
                os.makedirs(work_dir)
                print(f"Created working directory: {work_dir}")
            except Exception as e:
                print(f"Error creating working directory {work_dir}: {e}")
                return

    if not task_id:
        if prompt:
            short_hash = hashlib.md5(prompt.encode()).hexdigest()[:8]
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            task_id = f"{timestamp}_{short_hash}"
            print(f"Generated Task ID: {task_id}")
        else:
            print("Error: Provide a prompt or a task ID to resume/revert.")
            return

    role_tool_map = {}
    role_model_map = {}
    role_fallback_tool_map = {}
    role_fallback_model_map = {}

    for step in runtime_steps:
        # Tool
        attr_name_tool = step.arg_name.replace("-", "_")
        specific_tool_arg = getattr(args, attr_name_tool, None)
        if specific_tool_arg: role_tool_map[step.name] = specific_tool_arg
        elif step.tool: role_tool_map[step.name] = step.tool
        else: role_tool_map[step.name] = default_tool

        # Model
        attr_name_model = step.model_arg_name.replace("-", "_")
        specific_model_arg = getattr(args, attr_name_model, None)
        if specific_model_arg: role_model_map[step.name] = specific_model_arg
        elif step.model: role_model_map[step.name] = step.model
        else: role_model_map[step.name] = default_model

        # Fallback Tool
        attr_name_fb_tool = step.fallback_tool_arg_name.replace("-", "_")
        specific_fb_tool = getattr(args, attr_name_fb_tool, None)
        if specific_fb_tool: role_fallback_tool_map[step.name] = specific_fb_tool
        elif step.fallback_tool: role_fallback_tool_map[step.name] = step.fallback_tool

        # Fallback Model
        attr_name_fb_model = step.fallback_model_arg_name.replace("-", "_")
        specific_fb_model = getattr(args, attr_name_fb_model, None)
        if specific_fb_model: role_fallback_model_map[step.name] = specific_fb_model
        elif step.fallback_model: role_fallback_model_map[step.name] = step.fallback_model

    if args.tui:
        from .tui.app import SoftwareCompanyApp

        # Factory to create workflow instance injected with TUI adapter
        def workflow_factory(ui, task_id_override=None):
            return SoftwareCompanyWorkflow(
                task_id=task_id_override or task_id,
                steps=runtime_steps,
                state_directory=state_directory,
                role_tool_map=role_tool_map,
                role_model_map=role_model_map,
                role_fallback_tool_map=role_fallback_tool_map,
                role_fallback_model_map=role_fallback_model_map,
                default_tool=default_tool,
                work_dir=work_dir,
                max_workflow_iterations=max_workflow_iterations,
                default_timeout=default_timeout,
                ui=ui
            )

        app = SoftwareCompanyApp(workflow_factory=workflow_factory, state_directory=state_directory)
        
        # We need to manually inject user_request into state if present,
        # because the app runs workflow.run() without args or with hardcoded ones.
        
        # Alternate plan: We can manually prep the state file before launching TUI if it's a new task.
        if prompt:
             from .state_manager import StateManager
             # We need to init state manager to save the prompt.
             # This is a bit duplicative but safe.
             temp_sm = StateManager(task_id, state_directory)
             if temp_sm.get_context("user_request") is None:
                 temp_sm.set_context("user_request", prompt)
                 temp_sm.save_state()

        await app.run_async()

    else:
        workflow = SoftwareCompanyWorkflow(
            task_id=task_id,
            steps=runtime_steps,
            state_directory=state_directory,
            role_tool_map=role_tool_map,
            role_model_map=role_model_map,
            role_fallback_tool_map=role_fallback_tool_map,
            role_fallback_model_map=role_fallback_model_map,
            default_tool=default_tool,
            work_dir=work_dir,
            max_workflow_iterations=max_workflow_iterations,
            default_timeout=default_timeout
        )

        try:
            if args.revert:
                workflow.revert_last_step()
            elif args.run_step:
                try:
                    workflow.jump_to_step(args.run_step)
                    await workflow.run(prompt if prompt else "", single_step=args.step)
                except ValueError as e:
                    print(f"Error: {e}")
            else:
                await workflow.run(prompt if prompt else "", single_step=args.step)
        except asyncio.CancelledError:
            workflow.ui.stop_spinner()
            workflow.ui.print_warning("Operation cancelled by user.")

def main():
    try:
        asyncio.run(run_workflow())
    except KeyboardInterrupt:
        sys.exit(0)

if __name__ == "__main__":
    main()



import argparse
import sys
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from rich.theme import Theme
from rich.align import Align

LOGO = r"""
   _____       ______
  / ___/____  / __/ /__      ______  ________
  \__ \/ __ \/ /_/ __/ | /| / / __ \/ ___/ _ \
 ___/ / /_/ / __/ /_ | |/ |/ / /_/ / /  /  __/
/____/\____/_/  \__/ |__/|__/\__,_/_/   \___/

   ______
  / ____/___  ____ ___  ____  ____ _____  __  __
 / /   / __ \/ __ `__ \/ __ \/ __ `/ __ \/ / / /
/ /___/ /_/ / / / / / / /_/ / /_/ / / / / /_/ /
\____/\____/_/ /_/ /_/ .___/\__,_/_/ /_/\__, /
                    /_/                /____/
"""

class RichArgumentParser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.console = Console(theme=Theme({
            "info": "dim cyan",
            "warning": "magenta",
            "error": "bold red",
            "success": "bold green",
            "step": "bold blue",
            "json.key": "bold cyan",
            "json.str": "green",
            "highlight": "bold yellow",
            "role": "bold magenta",
            "tool": "bold cyan",
            "file": "underline blue",
        }))

    def print_help(self, file=None):
        if file is None:
            file = sys.stdout

        # 1. Logo
        self.console.print(Align.center(Text(LOGO, style="bold cyan")))

        # 2. Description
        if self.description:
            self.console.print(Align.center(Text(self.description, style="white")))
        self.console.print()

        # 3. Arguments Table (No Usage, No Categories)
        table = Table(show_header=True, header_style="bold magenta", box=None, expand=True, padding=(0, 2))
        table.add_column("Options/Arguments", style="bold cyan", ratio=1)
        table.add_column("Description", style="white", ratio=2)
        table.add_column("Default", style="dim", ratio=1)

        for group in self._action_groups:
            actions = [a for a in group._group_actions if a.help != argparse.SUPPRESS]
            if not actions:
                continue

            # Skip adding group title (Categories)

            for action in actions:
                flags, help_text, default = self._format_action_row(action)
                table.add_row(flags, help_text, default)

            # Optional: Add a subtle separator between groups if desired, but user asked to remove categories.
            # We'll just let them flow.

        self.console.print(table)
        self.console.print()

        # 4. Examples
        examples = """
[bold yellow]Examples:[/bold yellow]

  [dim]# Start a new task with a prompt[/dim]
  [green]software-company[/green] "Build a snake game in python"

  [dim]# Use a specific template[/dim]
  [green]software-company[/green] --template planning "Create a todo app"

  [dim]# List available tools[/dim]
  [green]software-company[/green] --list-tools

  [dim]# Resume a specific task[/dim]
  [green]software-company[/green] --id 20231027_100000_abc123

  [dim]# Revert the last step of a task[/dim]
  [green]software-company[/green] --revert --id 20231027_100000_abc123
        """
        self.console.print(Panel(Text.from_markup(examples.strip()), title="[bold]Quick Start[/bold]", border_style="cyan"))

    def _format_action_row(self, action):
        # Flags
        if action.option_strings:
            flags = ", ".join(action.option_strings)
            if action.nargs != 0 and action.dest:
                 metavar = action.metavar or action.dest.upper()
                 if action.nargs == '+':
                     flags += f" [{metavar} ...]"
                 elif action.nargs == '*':
                     flags += f" [{metavar} ...]"
                 elif action.nargs == '?':
                     flags += f" [{metavar}]"
                 else:
                     flags += f" {metavar}"

        else:
            # Positional
            flags = action.metavar or action.dest

        # Help
        help_text = action.help if action.help else ""

        # Default
        default = ""
        if action.default != argparse.SUPPRESS and action.default is not None:
             default = str(action.default)

        return flags, help_text, default



from typing import List, Optional
from .base import BaseCLIWrapper

class OpenCodeCLIWrapper(BaseCLIWrapper):
    """
    Wrapper for the OpenCode CLI tool.
    Uses 'opencode run' and '-c' for context continuation.
    """

    def __init__(self, command: str = "opencode", auto_approve: bool = True, model: Optional[str] = None, work_dir: Optional[str] = None, read_only: bool = False, sandbox_manager=None):
        super().__init__(command, auto_approve, model, work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
        self._has_started = False

    def _construct_command(self, prompt_text: str) -> List[str]:
        cmd_args = self.command.split() + ["run"]
        
        if self.model:
            cmd_args.extend(["-m", self.model])
        
        if self.session_id:
            cmd_args.extend(["-s", self.session_id])
        elif self._has_started:
            cmd_args.append("-c")
            
        cmd_args.append(prompt_text)
        return cmd_args

    def _parse_output(self, stdout: str) -> str:
        self._has_started = True
        return stdout.strip()


import os
from typing import List, Optional
from .base import BaseCLIWrapper
from .utils import strip_ansi_codes

class AiderCLIWrapper(BaseCLIWrapper):
    """
    Wrapper for the Aider CLI tool.
    """

    def __init__(self, command: str = "aider", auto_approve: bool = True, model: Optional[str] = None, work_dir: Optional[str] = None, read_only: bool = False, sandbox_manager=None):
        super().__init__(command, auto_approve, model, work_dir, read_only=read_only, sandbox_manager=sandbox_manager)

    def _construct_command(self, prompt_text: str) -> List[str]:
        cmd_args = self.command.split()

        # Non-interactive flags
        # We allow streaming to support real-time feedback (handled by BaseCLIWrapper)
        # cmd_args.append("--no-stream")
        cmd_args.append("--no-auto-commits")

        if self.auto_approve:
            cmd_args.append("--yes")

        if self.model:
            cmd_args.extend(["--model", self.model])

        if self.read_only:
             # If read-only, ensure we are strict in prompt.
             # Note: We still pass --yes to prevent blocking on confirmations,
             # but we rely on the prompt instructions to prevent edits.
             prompt_text = "READ-ONLY MODE: DO NOT EDIT FILES. " + prompt_text

        cmd_args.extend(["--message", prompt_text])
        return cmd_args

    def _parse_output(self, stdout: str) -> str:
        clean_output = strip_ansi_codes(stdout)
        return clean_output.strip()



from textual.widgets import Static, Header, Label, Button, ListView, ListItem
from textual.app import ComposeResult
from textual.reactive import reactive
from textual.containers import Vertical, Horizontal, VerticalScroll
from textual.message import Message
from rich.text import Text
import os
import json
from datetime import datetime


class TaskControlHeader(Static):
    """Custom header with a stylized logo."""
    
    def compose(self) -> ComposeResult:
        logo_text = Text(" ⚡ AUTO-SOFTWARE-CO ", style="bold italic #89b4fa on #181825")
        yield Label(logo_text, id="header-logo")
        yield Header(show_clock=True, id="main-header")


class TaskCard(Static):
    """Card widget for displaying task information."""
    
    class Selected(Message):
        """Posted when user selects a task."""
        def __init__(self, task_id: str):
            self.task_id = task_id
            super().__init__()
    
    class DeleteRequested(Message):
        """Posted when user requests to delete a task."""
        def __init__(self, task_id: str):
            self.task_id = task_id
            super().__init__()
    
    task_id = reactive("")
    status = reactive("Pending")
    current_step = reactive("Start")
    timestamp = reactive("")
    
    def __init__(self, task_id: str, status: str, current_step: str, timestamp: str, **kwargs):
        super().__init__(**kwargs)
        self.task_id = task_id
        self.status = status
        self.current_step = current_step
        self.timestamp = timestamp
        
        # Determine icon based on status
        if status == "Failed":
            self.icon = "❌"
        elif status == "Done":
            self.icon = "✔"
        else:
            self.icon = "⏳"
    
    def compose(self) -> ComposeResult:
        with Vertical(classes="task-card-container"):
            with Horizontal(classes="task-card-header"):
                yield Label(f"{self.icon} {self.task_id}", classes="task-card-title")
                delete_btn = Button("🗑", id=f"delete-task-{self.task_id}", classes="icon-btn delete-btn", variant="primary")
                yield delete_btn
            
            with Vertical(classes="task-card-body"):
                yield Label(f"Step: {self.current_step}", classes="task-card-info")
                yield Label(f"Time: {self.timestamp}", classes="task-card-info")
    
    def on_click(self):
        """Handle card click."""
        self.post_message(self.Selected(self.task_id))


class TaskExplorerSidebar(Static):
    """Sidebar widget to display tasks as cards."""
    
    class Selected(Message):
        def __init__(self, task_id: str):
            self.task_id = task_id
            super().__init__()
    
    class DeleteRequested(Message):
        """Posted when user requests to delete a task."""
        def __init__(self, task_id: str):
            self.task_id = task_id
            super().__init__()

    def __init__(self, state_directory: str, **kwargs):
        super().__init__(**kwargs)
        self.state_directory = state_directory
        self.tasks = []

    def compose(self) -> ComposeResult:
        yield Label("☰ PROJECT TASKS", classes="sidebar-header")
        yield VerticalScroll(id="task-cards-container")

    def on_mount(self):
        self.refresh_tasks()

    def refresh_tasks(self):
        container = self.query_one(VerticalScroll)
        container.remove_children()
        
        if not os.path.exists(self.state_directory):
            return

        tasks_data = []
        try:
            items = os.listdir(self.state_directory)
        except OSError:
            return

        for item in items:
            item_path = os.path.join(self.state_directory, item)
            if os.path.isdir(item_path):
                task_id = item
                state_file_path = os.path.join(item_path, f"task_{task_id}.json")

                if os.path.exists(state_file_path):
                    try:
                        mtime = os.path.getmtime(state_file_path)
                        with open(state_file_path, 'r') as f:
                            state = json.load(f)

                        timeline = state.get("timeline", [])
                        status = "Pending"
                        current_step = "Start"
                        
                        if timeline:
                            last = timeline[-1]
                            if last.get("status") == "FAILED":
                                status = "Failed"
                            elif last.get("status") == "COMPLETED":
                                status = "In Progress"
                            
                            current_step = last.get("step_name", "Unknown")

                        # Format timestamp
                        dt = datetime.fromtimestamp(mtime)
                        timestamp = dt.strftime("%Y-%m-%d %H:%M:%S")

                        tasks_data.append({
                            "id": task_id,
                            "status": status,
                            "current_step": current_step,
                            "timestamp": timestamp,
                            "mtime": mtime
                        })
                    except:
                        continue
        
        tasks_data.sort(key=lambda x: x["mtime"], reverse=True)

        for t in tasks_data:
            card = TaskCard(
                t['id'],
                t['status'],
                t['current_step'],
                t['timestamp'],
                id=f"task-card-{t['id']}",
                classes="task-card"
            )
            container.mount(card)

    def on_task_card_selected(self, message: TaskCard.Selected):
        """Handle task card selection."""
        self.post_message(self.Selected(message.task_id))
    
    def on_task_card_delete_requested(self, message: TaskCard.DeleteRequested):
        """Handle task deletion request."""
        self.post_message(self.DeleteRequested(message.task_id))


class StepCard(Static):
    """Card widget for displaying step information."""
    
    class RerunRequested(Message):
        """Posted when user requests to re-run a step."""
        def __init__(self, step_index: int):
            self.step_index = step_index
            super().__init__()
    
    role = reactive("")
    tool = reactive("Gemini")
    model = reactive("Default")
    status = reactive("Pending")
    duration = reactive("-")
    
    def __init__(self, step_index: int, step_data, **kwargs):
        super().__init__(**kwargs)
        self.step_index = step_index
        self.step_data = step_data
        self.role = step_data.name
        self.tool = step_data.tool or "Default"
        self.model = step_data.model or "Default"
        
        input_list = step_data.input_keys or []
        self.input_keys = ", ".join(input_list) if input_list else "None"
        
        output_list = step_data.output_keys or []
        self.output_keys = ", ".join(output_list) if output_list else "None"
        
        self.output_content_map = {}
        self.generated_files = []
        self.output_visible = False  # Track if output is expanded

    def compose(self) -> ComposeResult:
        with Horizontal(classes="step-header"):
            yield Label(f"{self.step_index + 1}. {self.role}", classes="step-title")
            yield Label(self.status, classes="step-status")
            rerun_btn = Button("🔄", id=f"rerun-btn-{self.step_index}", classes="icon-btn rerun-btn", variant="primary")
            yield rerun_btn
            # Toggle output button
            toggle_btn = Button("▼", id=f"toggle-output-{self.step_index}", classes="icon-btn toggle-btn", variant="primary")
            yield toggle_btn
        
        with Vertical(classes="step-body"):
            yield Label(f"Tool: {self.tool} ({self.model})", classes="step-meta")
            yield Label(f"In: {self.input_keys}", classes="step-io")
            yield Label(f"Out: {self.output_keys}", classes="step-io")
            yield Vertical(classes="file-list-container")
            # Output section (hidden by default)
            yield Vertical(id=f"output-section-{self.step_index}", classes="output-section hidden")
        
        with Horizontal(classes="step-footer"):
            yield Label("", classes="step-output-files")
            yield Label(f"{self.duration}", classes="step-time")

    def watch_status(self, val):
        if not self.is_mounted:
            return
        lbl = self.query_one(".step-status", Label)
        
        if val == "Running":
            lbl.update(f"🔄 Running")
            self.add_class("running")
        elif val == "Completed":
            lbl.update(f"✔ Done")
            self.remove_class("running")
            self.add_class("completed")
        elif val == "Failed":
            lbl.update(f"❌ Failed")
            self.add_class("failed")
        else:
            lbl.update(f"○ Pending")

    def watch_duration(self, val):
        if not self.is_mounted:
            return
        self.query_one(".step-time", Label).update(f"{val}")

    def set_output(self, output_dict):
        """Stores the full output dictionary."""
        self.output_content_map = {k: str(v) if not isinstance(v, str) else v for k, v in output_dict.items()}

    def set_files(self, files: list):
        """Populates the file list with copy buttons."""
        if not self.is_mounted:
            return
        
        self.generated_files = files
        container = self.query_one(".file-list-container", Vertical)
        container.remove_children()
        
        if files:
            for key, path in files:
                row = Horizontal(classes="file-row")
                filename = os.path.basename(path)
                label = Label(f"📄 {filename}", classes="file-path")
                btn_id = f"copy-btn-{self.step_index}-{key}"
                btn = Button("📋", id=btn_id, classes="icon-btn", variant="primary")
                btn.copy_key = key
                
                container.mount(row)
                row.mount(label)
                row.mount(btn)
        
        self._update_footer_files(files)

    def on_button_pressed(self, event: Button.Pressed):
        """Handle button clicks: copy, re-run, and toggle buttons."""
        # Handle toggle output button
        if event.button.id and event.button.id.startswith("toggle-output-"):
            self._toggle_output()
            return
        
        # Handle re-run button
        if event.button.id and event.button.id.startswith("rerun-btn-"):
            self.post_message(self.RerunRequested(self.step_index))
            self.app.notify(f"🔄 Re-running {self.role}...", timeout=2)
            return
        
        # Handle copy button clicks for step output
        if hasattr(event.button, "copy_key"):
            key = event.button.copy_key
            content = self.output_content_map.get(key)
            if content:
                try:
                    import pyperclip
                    pyperclip.copy(content)
                    self.app.notify(f"✓ Copied {key} to clipboard!", timeout=2)
                except Exception as e:
                    try:
                        self.app.copy_to_clipboard(content)
                        self.app.notify(f"✓ Copied {key} to clipboard!", timeout=2)
                    except Exception as e2:
                        self.app.notify(f"Failed to copy: {e2}", severity="warning", timeout=2)
            else:
                self.app.notify(f"No content for {key}", severity="warning", timeout=2)

    def _update_footer_files(self, files: list):
        """Update the footer label with output file paths."""
        if not self.is_mounted:
            return
        
        try:
            footer_label = self.query_one(".step-output-files", Label)
            if files:
                file_paths = []
                for key, path in files:
                    filename = os.path.basename(path)
                    file_paths.append(filename)
                file_info = " | ".join(file_paths)
                footer_label.update(file_info)
            else:
                footer_label.update("")
        except:
            pass
    
    def _toggle_output(self):
        """Toggle output section visibility."""
        try:
            output_section = self.query_one(f"#output-section-{self.step_index}", Vertical)
            toggle_btn = self.query_one(f"#toggle-output-{self.step_index}", Button)
            
            if self.output_visible:
                # Hide output
                output_section.add_class("hidden")
                toggle_btn.label = "▼"
                self.output_visible = False
            else:
                # Show output
                output_section.remove_class("hidden")
                toggle_btn.label = "▲"
                self.output_visible = True
                # Populate output if not already done
                if not output_section.children:
                    self._populate_output_section(output_section)
        except Exception as e:
            self.app.notify(f"Error toggling output: {e}", severity="warning", timeout=2)
    
    def _populate_output_section(self, container: Vertical):
        """Populate the output section with step outputs."""
        for key, content in self.output_content_map.items():
            # Add key label
            key_label = Label(f"[bold]{key}:[/bold]", classes="output-key")
            container.mount(key_label)
            # Add content (selectable text)
            content_label = Label(str(content), classes="output-content", expand=True)
            container.mount(content_label)



import asyncio
from typing import Sequence, Type, Optional, Callable
from autogen_agentchat.agents import BaseChatAgent
from autogen_agentchat.base import Response
from autogen_agentchat.messages import BaseChatMessage, TextMessage
from autogen_core import CancellationToken

from .cli.base import BaseCLIWrapper
from .cli.gemini import GeminiCLIWrapper
from .cli.amazonq import AmazonQCLIWrapper
from .cli.opencode import OpenCodeCLIWrapper
from .cli.aider import AiderCLIWrapper
from .cli.github_copilot import GitHubCopilotCLIWrapper

class CLIAgent(BaseChatAgent):
    """
    An AutoGen agent that delegates tasks to a CLI tool wrapper.
    """
    
    def __init__(self, name: str, wrapper: BaseCLIWrapper, description: str = "A CLI-based agent.", timeout: Optional[int] = None):
        super().__init__(name, description)
        self.wrapper = wrapper
        self.timeout = timeout
        self.stream_callback: Optional[Callable[[str, str], None]] = None

    def set_stream_callback(self, callback: Callable[[str, str], None]):
        self.stream_callback = callback

    @property
    def produced_message_types(self) -> Sequence[Type[BaseChatMessage]]:
        return (TextMessage,)

    async def on_messages(
        self,
        messages: Sequence[BaseChatMessage],
        cancellation_token: CancellationToken,
    ) -> Response:
        """
        Handles incoming messages by passing the last message content to the CLI wrapper.
        """
        if not messages:
            return Response(chat_message=TextMessage(content="", source=self.name))

        # Extract last message content
        last_msg = messages[-1]
        user_input = last_msg.content if isinstance(last_msg, TextMessage) else str(last_msg)
        
        # Run CLI wrapper in a thread to avoid blocking the event loop
        # since subprocess is blocking.
        if self.timeout:
            response_text = await asyncio.to_thread(
                self.wrapper.ask,
                user_input,
                timeout=self.timeout,
                stream_callback=self.stream_callback
            )
        else:
            response_text = await asyncio.to_thread(
                self.wrapper.ask,
                user_input,
                stream_callback=self.stream_callback
            )
        
        return Response(chat_message=TextMessage(content=response_text, source=self.name))

    async def on_reset(self, cancellation_token: CancellationToken) -> None:
        """
        Resets the agent state (clears CLI session).
        """
        self.wrapper.close()

def create_gemini_agent(name: str = "gemini_agent", model: str = None, work_dir: str = None, timeout: int = None, read_only: bool = False, sandbox_manager=None) -> CLIAgent:
    wrapper = GeminiCLIWrapper(model=model, work_dir=work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
    return CLIAgent(name, wrapper, description="Agent powered by Gemini CLI", timeout=timeout)

def create_amazonq_agent(name: str = "q_agent", model: str = None, work_dir: str = None, timeout: int = None, read_only: bool = False, sandbox_manager=None) -> CLIAgent:
    wrapper = AmazonQCLIWrapper(model=model, work_dir=work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
    return CLIAgent(name, wrapper, description="Agent powered by Amazon Q CLI", timeout=timeout)

def create_opencode_agent(name: str = "opencode_agent", model: str = None, work_dir: str = None, timeout: int = None, read_only: bool = False, sandbox_manager=None) -> CLIAgent:
    wrapper = OpenCodeCLIWrapper(model=model, work_dir=work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
    return CLIAgent(name, wrapper, description="Agent powered by OpenCode CLI", timeout=timeout)

def create_aider_agent(name: str = "aider_agent", model: str = None, work_dir: str = None, timeout: int = None, read_only: bool = False, sandbox_manager=None) -> CLIAgent:
    wrapper = AiderCLIWrapper(model=model, work_dir=work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
    return CLIAgent(name, wrapper, description="Agent powered by Aider CLI", timeout=timeout)

def create_github_copilot_agent(name: str = "copilot_agent", model: str = None, work_dir: str = None, timeout: int = None, read_only: bool = False, sandbox_manager=None) -> CLIAgent:
    wrapper = GitHubCopilotCLIWrapper(model=model, work_dir=work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
    return CLIAgent(name, wrapper, description="Agent powered by GitHub Copilot CLI", timeout=timeout)



import re

def strip_ansi_codes(text: str) -> str:
    """
    Removes ANSI escape sequences (colors, bold text) often used by CLI tools.
    This ensures the output we capture is clean text.
    """
    ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
    return ansi_escape.sub("", text)




from textual.app import App
from typing import List, Optional
import threading
from ..ui import UserInterface

class TuiUIAdapter(UserInterface):
    def __init__(self, app: App):
        self.app = app

    def _safe_call(self, callback, *args, **kwargs):
        if threading.get_ident() == self.app._thread_id:
            callback(*args, **kwargs)
        else:
            self.app.call_from_thread(lambda: callback(*args, **kwargs))

    def print_step_header(self, step_number: int, step_name: str, work_dir: str = None):
        self._safe_call(self.app.on_step_start, step_number, step_name)

    def print_output(self, step_name: str, output: dict):
        self._safe_call(self.app.update_step_output, step_name, output)

    # Implement other methods to avoid crashes, but they might be no-ops or simple notifications
    def print_assignment(self, role: str, tool: str, model: str):
        pass 

    def print_timeout(self, timeout_val: int):
        pass

    def print_input_loading(self, input_key: str, path: str):
        pass

    def print_tool_start(self, step_name: str, tool_name: str):
        pass

    def start_spinner(self, message: str):
        pass

    def stop_spinner(self):
        pass

    def print_error(self, message: str):
        self._safe_call(self.app.notify, message, severity="error")

    def print_warning(self, message: str):
        self._safe_call(self.app.notify, message, severity="warning")

    def print_info(self, message: str):
        self._safe_call(self.app.write_to_console, f"[dim]{message}[/dim]")

    def print_success(self, message: str):
        self._safe_call(self.app.write_to_console, f"[bold green]{message}[/bold green]")

    def print_stream_line(self, line: str, source: str = "stdout"):
        self._safe_call(self.app.stream_log, line)

    def input(self, prompt: str) -> str:
        return ""

    def print_user_intervention(self, step_name: str, files: list):
        self._safe_call(self.app.notify, f"User Intervention Required: {step_name}", severity="warning")
        self._safe_call(self.app.write_to_console, f"[bold yellow]USER INTERVENTION REQUIRED for {step_name}[/bold yellow]")
        for k, p in files:
            self._safe_call(self.app.write_to_console, f" - {k}: {p}")

    def print_artifacts(self, step_name: str, artifacts: list):
        self._safe_call(self.app.update_step_artifacts, step_name, artifacts)

    def print_step_outputs_summary(self, step_name: str, artifacts: list):
        self._safe_call(self.app.update_step_artifacts, step_name, artifacts)

    def print_newline(self):
        self._safe_call(self.app.write_to_console, "")



import re
from typing import List, Optional, Callable
from .base import BaseCLIWrapper
from .utils import strip_ansi_codes

class AmazonQCLIWrapper(BaseCLIWrapper):
    """
    Wrapper for the Amazon Q CLI (typically 'q').
    Supports session resumption via CWD context and auto-approval of tools.
    """

    def __init__(self, command: str = "q", auto_approve: bool = True, model: Optional[str] = None, work_dir: Optional[str] = None, read_only: bool = False, sandbox_manager=None):
        super().__init__(command, auto_approve, model, work_dir, read_only=read_only, sandbox_manager=sandbox_manager)
        # We track if a session has started to know when to pass -r
        self._has_started = False

    def _construct_command(self, prompt_text: str) -> List[str]:
        cmd_args = self.command.split()
        cmd_args.append("chat")
        cmd_args.append("--no-interactive")
        cmd_args.append("--wrap")
        cmd_args.append("never")
        
        if self.auto_approve and not self.read_only:
            cmd_args.append("--trust-all-tools")
            
        if self.model:
            cmd_args.extend(["--model", self.model])
            
        if self._has_started:
            cmd_args.append("-r")
            
        cmd_args.append(prompt_text)
        return cmd_args

    def _process_stream_line(self, line: str, source: str, callback: Callable[[str, str], None]) -> None:
        """
        Strip ANSI codes from streamed output for cleaner display.
        """
        clean_line = strip_ansi_codes(line)
        callback(clean_line, source)

    def _parse_output(self, stdout: str) -> str:
        self._has_started = True
        
        # Strip ANSI codes first!
        clean_stdout = strip_ansi_codes(stdout)

        separator_regex = re.compile(r"━{10,}")
        parts = separator_regex.split(clean_stdout)
        
        if len(parts) > 1:
            content = parts[-1]
        else:
            content = clean_stdout

        lines = content.splitlines()
        cleaned_lines = []
        
        ignore_patterns = [
            re.compile(r"^🤖 You are chatting with"),
            re.compile(r"^Checkpoints are not available"),
            re.compile(r"^Picking up where we left off"),
        ]
        
        for line in lines:
            line = line.strip()
            if not line:
                continue
                
            should_ignore = False
            for pattern in ignore_patterns:
                if pattern.search(line):
                    should_ignore = True
                    break
            
            if not should_ignore:
                if line.startswith("> "):
                    line = line[2:]
                cleaned_lines.append(line)

        return "\n".join(cleaned_lines).strip()


from textual.app import ComposeResult
from textual.screen import Screen
from textual.widgets import Header, Footer, DataTable, Button, Label
from textual.containers import Container, Vertical
from textual.binding import Binding
import os
import json
from datetime import datetime

class TaskSelectionScreen(Screen):
    """Screen for exploring and selecting existing tasks."""

    BINDINGS = [
        Binding("escape", "app.pop_screen", "Cancel"),
    ]

    CSS = """
    TaskSelectionScreen {
        align: center middle;
    }

    #dialog {
        grid-size: 2;
        grid-gutter: 1 2;
        grid-rows: 1fr 3;
        padding: 0 1;
        width: 80%;
        height: 80%;
        border: thick $background 80%;
        background: $surface;
    }

    #task-table {
        height: 1fr;
        border: solid #89b4fa;
    }
    
    #title-label {
        text-style: bold;
        color: #cba6f7;
        margin-bottom: 1;
        text-align: center;
    }
    """

    def __init__(self, state_directory: str, **kwargs):
        super().__init__(**kwargs)
        self.state_directory = state_directory
        self.selected_task_id = None

    def compose(self) -> ComposeResult:
        with Vertical(id="dialog"):
            yield Label("Select a Task to Resume", id="title-label")
            yield DataTable(id="task-table", cursor_type="row")
            yield Button("Cancel", variant="error", id="cancel-btn")

    def on_mount(self):
        table = self.query_one(DataTable)
        table.add_columns("Task ID", "Status", "Last Updated", "Description")
        self.load_tasks()

    def load_tasks(self):
        table = self.query_one(DataTable)
        table.clear()

        if not os.path.exists(self.state_directory):
            return

        tasks = []
        try:
            items = os.listdir(self.state_directory)
        except OSError:
            return

        for item in items:
            item_path = os.path.join(self.state_directory, item)
            if os.path.isdir(item_path):
                task_id = item
                state_file_path = os.path.join(item_path, f"task_{task_id}.json")

                if os.path.exists(state_file_path):
                    try:
                        mtime = os.path.getmtime(state_file_path)
                        last_updated = datetime.fromtimestamp(mtime).strftime('%Y-%m-%d %H:%M')

                        with open(state_file_path, 'r') as f:
                            state = json.load(f)

                        context = state.get("context", {})
                        user_request = context.get("user_request", "No description.")
                        if isinstance(user_request, dict):
                            description = "[File Reference]"
                        else:
                            description = str(user_request).replace('\n', ' ')
                            if len(description) > 60:
                                description = description[:57] + "..."

                        timeline = state.get("timeline", [])
                        status = "Pending"
                        if timeline:
                            last_entry = timeline[-1]
                            if last_entry.get("status") == "FAILED":
                                status = "Failed"
                            elif last_entry.get("status") == "COMPLETED":
                                # Check if it was the last step
                                status = "In Progress" # Simplified logic

                        tasks.append({
                            "id": task_id,
                            "status": status,
                            "description": description,
                            "last_updated": last_updated,
                            "mtime": mtime
                        })
                    except:
                        continue

        tasks.sort(key=lambda x: x["mtime"], reverse=True)

        for task in tasks:
            table.add_row(
                task["id"], 
                task["status"], 
                task["last_updated"], 
                task["description"],
                key=task["id"] # Use ID as row key
            )
        
        table.focus()

    def on_data_table_row_selected(self, event: DataTable.RowSelected):
        self.selected_task_id = event.row_key.value
        self.dismiss(self.selected_task_id)

    def on_button_pressed(self, event: Button.Pressed):
        if event.button.id == "cancel-btn":
            self.dismiss(None)



import os
import json
import time
from datetime import datetime
from typing import List

def list_tools(tool_names: List[str]):
    """Lists available tool wrappers."""
    print("Available Tools:")
    for tool in tool_names:
        print(f"  - {tool}")

def list_tasks(state_directory: str):
    """Lists tasks in the state directory."""
    if not os.path.exists(state_directory):
        print(f"State directory '{state_directory}' does not exist.")
        return

    tasks = []

    try:
        items = os.listdir(state_directory)
    except OSError as e:
        print(f"Error accessing state directory: {e}")
        return

    for item in items:
        item_path = os.path.join(state_directory, item)
        if os.path.isdir(item_path):
            task_id = item
            state_file_name = f"task_{task_id}.json"
            state_file_path = os.path.join(item_path, state_file_name)

            if os.path.exists(state_file_path):
                try:
                    # Get modification time
                    mtime = os.path.getmtime(state_file_path)
                    last_updated = datetime.fromtimestamp(mtime).strftime('%Y-%m-%d %H:%M:%S')

                    with open(state_file_path, 'r') as f:
                        state = json.load(f)

                    context = state.get("context", {})
                    user_request = context.get("user_request", "No description provided.")

                    # Handle if user_request is a file reference (unlikely but possible)
                    if isinstance(user_request, dict) and user_request.get("type") == "file":
                         # We don't want to read the file here, just show a placeholder
                         description = f"[File: {user_request.get('path')}]"
                    else:
                        description = str(user_request).replace('\n', ' ')

                    if len(description) > 50:
                        description = description[:47] + "..."

                    timeline = state.get("timeline", [])
                    status = "Pending"
                    if timeline:
                        last_entry = timeline[-1]
                        if last_entry.get("status") == "FAILED":
                            status = "Failed"
                        else:
                            status = "In Progress"

                    tasks.append({
                        "id": task_id,
                        "status": status,
                        "description": description,
                        "last_updated": last_updated,
                        "mtime": mtime
                    })

                except Exception:
                    # Skip malformed tasks
                    continue

    if not tasks:
        print("No previous tasks found.")
        return

    # Sort by last updated (descending)
    tasks.sort(key=lambda x: x["mtime"], reverse=True)

    headers = ["Task ID", "Status", "Last Updated", "Description"]
    widths = [len(h) for h in headers]

    for task in tasks:
        widths[0] = max(widths[0], len(task["id"]))
        widths[1] = max(widths[1], len(task["status"]))
        widths[2] = max(widths[2], len(task["last_updated"]))
        widths[3] = max(widths[3], len(task["description"]))

    fmt = f"{{:<{widths[0]}}}  {{:<{widths[1]}}}  {{:<{widths[2]}}}  {{:<{widths[3]}}}"

    print(fmt.format(*headers))
    print("-" * (sum(widths) + 6))

    for task in tasks:
        print(fmt.format(task["id"], task["status"], task["last_updated"], task["description"]))



from textual.app import App, ComposeResult
from textual.widgets import Header, Footer, RichLog, Label, Markdown
from textual.containers import Container, Horizontal, Vertical, VerticalScroll
from textual.binding import Binding
from datetime import datetime
import asyncio
import json
import os
import shutil

from .widgets import TaskControlHeader, TaskExplorerSidebar, StepCard
from .adapter import TuiUIAdapter
from ..workflow import SoftwareCompanyWorkflow

class SoftwareCompanyApp(App):
    TITLE = "Task Control"
    SUB_TITLE = "Autonomous Agent Orchestration"
    
    BINDINGS = [
        Binding("q", "quit", "✖ Quit", show=True),
        Binding("s", "stop_workflow", "⏹ Stop", show=True),
        Binding("b", "toggle_sidebar", "☰ Sidebar", show=True),
        Binding("d", "toggle_dark", "🌗 Theme", show=True)
    ]

    CSS = """
    /* --- Theming --- */
    
    /* Global Backgrounds & Colors */
    Screen { background: #1e1e2e; color: #cdd6f4; }
    .-light-mode Screen { background: #eff1f5; color: #4c4f69; }

    TaskControlHeader {
        dock: top;
        background: #11111b;
        color: #89b4fa;
        height: 3;
        layout: horizontal;
        border-bottom: solid #313244;
    }
    .-light-mode TaskControlHeader { background: #dce0e8; color: #1e66f5; border-bottom: solid #bcc0cc; }

    #header-logo {
        width: auto;
        height: 3;
        content-align: center middle;
        padding: 0 2;
        background: #181825;
        color: #89b4fa;
        text-style: bold italic;
    }
    .-light-mode #header-logo { background: #e6e9ef; color: #1e66f5; }

    #main-header {
        width: 1fr;
        background: #11111b;
        color: #cdd6f4;
    }
    .-light-mode #main-header { background: #dce0e8; color: #4c4f69; }

    Footer {
        dock: bottom;
        background: #11111b;
        color: #a6e3a1;
        height: 1;
    }
    .-light-mode Footer { background: #dce0e8; color: #40a02b; }

    /* Layout */
    #main-layout { width: 100%; height: 1fr; }

    /* Sidebar */
    #sidebar {
        width: 30%;
        height: 100%;
        background: #181825;
        padding: 1;
        transition: width 200ms in_out_cubic;
        border-right: solid #313244;
    }
    .-light-mode #sidebar { background: #e6e9ef; border-right: solid #bcc0cc; }
    
    #sidebar.hidden { display: none; width: 0%; border: none; }
    
    .sidebar-header {
        text-align: center;
        color: #89b4fa;
        text-style: bold;
        padding-bottom: 1;
        border-bottom: solid #313244;
    }
    .-light-mode .sidebar-header { color: #1e66f5; border-bottom: solid #bcc0cc; }

    ListView { background: #181825; border: none; }
    .-light-mode ListView { background: #e6e9ef; }
    
    ListItem { padding: 1; background: #181825; color: #cdd6f4; }
    .-light-mode ListItem { background: #e6e9ef; color: #4c4f69; }
    
    ListItem:hover { background: #313244; }
    .-light-mode ListItem:hover { background: #bcc0cc; }
    
    ListItem.--highlight { background: #313244; color: #89b4fa; }
    .-light-mode ListItem.--highlight { background: #bcc0cc; color: #1e66f5; }

    /* Main Content Area */
    #right-side-container { width: 1fr; height: 100%; background: #1e1e2e; }
    .-light-mode #right-side-container { background: #eff1f5; }

    #main-content { width: 100%; height: 1fr; padding: 1 2; }
    
    #task-header {
        height: 3;
        background: #1e1e2e;
        color: #bac2de;
        text-style: bold;
        content-align: center middle;
        border-bottom: solid #313244;
        margin-bottom: 1;
    }
    .-light-mode #task-header { background: #eff1f5; color: #6c6f85; border-bottom: solid #bcc0cc; }

    #steps-container { height: 1fr; scrollbar-background: #1e1e2e; scrollbar-color: #313244; }
    .-light-mode #steps-container { scrollbar-background: #eff1f5; scrollbar-color: #bcc0cc; }

    #bottom-deck {
        width: 100%;
        height: 12;
        background: #11111b;
        border-top: solid #313244;
    }
    .-light-mode #bottom-deck { background: #dce0e8; border-top: solid #bcc0cc; }

    #log-panel {
        width: 100%;
        height: 1fr;
        background: #11111b;
        color: #cdd6f4;
        border: none;
    }
    .-light-mode #log-panel { background: #dce0e8; color: #4c4f69; }
    
    #log-label {
        background: #313244;
        color: #cdd6f4;
        padding-left: 1;
        text-style: bold;
    }
    .-light-mode #log-label { background: #bcc0cc; color: #4c4f69; }

    /* Step Card Styles */
    StepCard {
        height: auto;
        background: #181825;
        margin-bottom: 1;
        padding: 1 1;
        border-left: thick #bac2de;
    }
    .-light-mode StepCard { background: #e6e9ef; border-left: thick #9399b2; }
    
    StepCard.running { border-left: thick #89b4fa; background: #313244; }
    .-light-mode StepCard.running { border-left: thick #1e66f5; background: #ccd0da; }
    
    StepCard.completed { border-left: thick #a6e3a1; }
    .-light-mode StepCard.completed { border-left: thick #40a02b; }
    
    StepCard.failed { border-left: thick #f38ba8; }
    .-light-mode StepCard.failed { border-left: thick #d20f39; }

    .step-header { height: auto; margin: 0; padding: 0; }
    .step-title { width: 1fr; text-style: bold; color: #f9e2af; }
    .-light-mode .step-title { color: #df8e1d; }
    
    .step-status { width: auto; color: #bac2de; }
    .-light-mode .step-status { color: #6c6f85; }

    .step-body { height: auto; margin: 0; padding: 0; }
    .step-meta { color: #89b4fa; }
    .-light-mode .step-meta { color: #1e66f5; }
    
    .step-io { color: #9399b2; text-style: italic; }
    .-light-mode .step-io { color: #6c6f85; }

    .file-list-container { height: auto; margin: 0; padding: 0; }
    .file-row { height: auto; margin: 0; padding: 0; }
    .file-path { color: #a6e3a1; width: 1fr; }
    .-light-mode .file-path { color: #40a02b; }

    .icon-btn {
        width: 4; min-width: 4; height: 1; border: none;
        background: transparent; color: #89b4fa; padding: 0; margin: 0;
    }
    .-light-mode .icon-btn { color: #1e66f5; }
    
    .rerun-btn { color: #f9e2af; }
    .-light-mode .rerun-btn { color: #df8e1d; }
    
    .toggle-btn { color: #cdd6f4; }
    .-light-mode .toggle-btn { color: #4c4f69; }
    
    .output-section { height: auto; padding: 1 2; border-top: solid #313244; margin-top: 1; }
    .-light-mode .output-section { border-top: solid #bcc0cc; }
    
    .output-section.hidden { display: none; }
    
    .output-key { color: #f9e2af; text-style: bold; margin-top: 1; }
    .-light-mode .output-key { color: #df8e1d; }
    
    .output-content { color: #a6e3a1; width: 100%; margin: 0 1; }
    .-light-mode .output-content { color: #40a02b; }
    
    .task-card { height: auto; margin-bottom: 1; }
    .task-card-container { height: auto; }
    .task-card-header { height: auto; width: 100%; }
    .task-card-body { height: auto; }
    .task-card-info { height: auto; }
    
    .delete-btn { color: #f38ba8; }
    .-light-mode .delete-btn { color: #d20f39; }
    
    .task-list-item { height: auto; padding: 1; margin: 0; }
    .task-list-item Horizontal { width: 100%; }
    
    .step-footer { height: auto; margin-top: 1; padding: 0; border-top: solid #313244; }
    .-light-mode .step-footer { border-top: solid #bcc0cc; }
    
    .step-output-files { width: 1fr; color: #a6e3a1; text-style: dim; }
    .-light-mode .step-output-files { color: #40a02b; }
    
    .step-time { width: auto; color: #bac2de; text-align: right; }
    .-light-mode .step-time { color: #6c6f85; }
    """

    def __init__(self, workflow_factory, state_directory: str, **kwargs):
        super().__init__(**kwargs)
        self._workflow_factory = workflow_factory
        self.state_directory = state_directory
        self.workflow = None
        self._worker = None
        self._step_start_times = {}

    def compose(self) -> ComposeResult:
        yield TaskControlHeader()
        
        with Horizontal(id="main-layout"):
            # Sidebar
            with Vertical(id="sidebar"):
                yield TaskExplorerSidebar(self.state_directory)
            
            # Right Side (Stage + Logs)
            with Vertical(id="right-side-container"):
                # Main Content
                with Vertical(id="main-content"):
                    yield Label("Task Details", id="task-header")
                    yield VerticalScroll(id="steps-container")
                
                # Bottom Deck
                with Vertical(id="bottom-deck"):
                    yield Label("EXECUTION LOGS", id="log-label")
                    yield RichLog(highlight=True, markup=True, id="log-panel", wrap=True)
        
        yield Footer()

    async def on_mount(self):
        # Initial Load using default args (from CLI)
        await self.load_workflow_async()

    async def on_task_explorer_sidebar_selected(self, message: TaskExplorerSidebar.Selected):
        await self.load_workflow_async(task_id_override=message.task_id)
    
    async def on_task_explorer_sidebar_delete_requested(self, message: TaskExplorerSidebar.DeleteRequested):
        """Handle task deletion request."""
        import shutil
        task_dir = os.path.join(self.state_directory, message.task_id)
        try:
            shutil.rmtree(task_dir)
            self.notify(f"✓ Task {message.task_id} deleted!", timeout=2)
            # Refresh task list
            sidebar = self.query_one(TaskExplorerSidebar)
            sidebar.refresh_tasks()
        except Exception as e:
            self.notify(f"Failed to delete task: {e}", severity="error", timeout=2)
    
    async def on_step_card_rerun_requested(self, message: StepCard.RerunRequested):
        """Handle step re-run request."""
        if not self.workflow:
            self.notify("No workflow loaded", severity="warning")
            return
        
        step = self.workflow.steps[message.step_index]
        self.notify(f"🔄 Re-running step: {step.name}...", timeout=1)
        
        # Revert the step to allow re-execution
        self.workflow.revert_last_step()
        
        # Jump to the specific step
        self.workflow.jump_to_step(step.name)
        
        # Start workflow execution from that step
        if self._worker and self._worker.is_running:
            self._worker.cancel()
        
        self._worker = self.run_worker(self._run_workflow())

    async def load_workflow_async(self, task_id_override=None):
        if self._worker and self._worker.is_running:
            self._worker.cancel()

        try:
            self.query_one("#log-panel", RichLog).clear()
        except:
            pass

        adapter = TuiUIAdapter(self)
        self.workflow = self._workflow_factory(ui=adapter, task_id_override=task_id_override)
        
        self.query_one("#task-header", Label).update(f"Task ID: {self.workflow.task_id}")

        container = self.query_one("#steps-container", VerticalScroll)
        await container.remove_children()
        
        # Load history
        timeline = self.workflow.state_manager.get_timeline()
        context = self.workflow.state_manager.state.get("context", {})
        
        for i, step in enumerate(self.workflow.steps):
            card = StepCard(i, step, id=f"step-card-{i}")
            await container.mount(card)
            
            # hydration
            step_entry = next((e for e in timeline if e["step_name"] == step.name and e["status"] == "COMPLETED"), None)
            
            if step_entry:
                card.status = "Completed"
                card.duration = f"{step_entry.get('duration', '-')}"
                
                outputs = {}
                files = []
                for key in step.output_keys:
                    val = context.get(key)
                    if val:
                        outputs[key] = val
                        if isinstance(val, dict) and val.get("type") == "file":
                            files.append((key, val["path"]))
                
                if outputs:
                    card.set_output(outputs)
                if files:
                    card.set_files(files)

        self._worker = self.run_worker(self.run_workflow_task, exclusive=True)

    async def run_workflow_task(self):
        try:
            await self.workflow.run()
        except asyncio.CancelledError:
             pass
        except Exception as e:
            self.notify(f"Workflow Error: {e}", severity="error")

    def action_stop_workflow(self):
        if self._worker and self._worker.is_running:
            self._worker.cancel()
            self.notify("Workflow Stopped")

    def action_toggle_sidebar(self):
        sidebar = self.query_one("#sidebar")
        sidebar.toggle_class("hidden")

    def action_load_mission(self):
        def on_mission_selected(task_id):
            if task_id:
                self.write_to_console(f"[bold cyan]Switching to task: {task_id}[/bold cyan]")
                self.call_from_thread(self.load_workflow_async, task_id_override=task_id)

        self.push_screen(TaskSelectionScreen(self.state_directory), on_mission_selected)

    def action_toggle_dark(self):
        self.toggle_class("-light-mode")

    # --- Adapter Methods ---

    def on_step_start(self, step_number: int, step_name: str):
        idx = step_number - 1
        try:
            card = self.query_one(f"#step-card-{idx}", StepCard)
            card.status = "Running"
            self._step_start_times[idx] = datetime.now()
            card.scroll_visible()
        except:
            pass

    def update_step_output(self, step_name: str, output: dict):
        for i, step in enumerate(self.workflow.steps):
            if step.name == step_name:
                try:
                    card = self.query_one(f"#step-card-{i}", StepCard)
                    card.status = "Completed"
                    card.set_output(output)
                    
                    if i in self._step_start_times:
                        delta = datetime.now() - self._step_start_times[i]
                        card.duration = f"{delta.total_seconds():.1f}s"
                except:
                    pass
                break

    def update_step_artifacts(self, step_name: str, artifacts: list):
        for i, step in enumerate(self.workflow.steps):
            if step.name == step_name:
                try:
                    card = self.query_one(f"#step-card-{i}", StepCard)
                    card.set_files(artifacts)
                except Exception as e:
                    self.write_to_console(f"[bold red]Error updating artifacts for {step_name}: {e}[/bold red]")
                break

    def write_to_console(self, message: str):
        try:
            log_panel = self.query_one("#log-panel", RichLog)
            log_panel.write(message)
        except:
            pass

    def stream_log(self, text: str):
        try:
            log_panel = self.query_one("#log-panel", RichLog)
            log_panel.write(text)
        except:
            pass


import asyncio
import json
import os
from datetime import datetime
from typing import List, Dict, Any, Optional
from dataclasses import dataclass

from autogen_agentchat.messages import TextMessage
from .agent_adapter import create_gemini_agent, create_amazonq_agent, create_opencode_agent, create_aider_agent, create_github_copilot_agent, CLIAgent
from .config import load_config, DEFAULT_CONFIG_FILENAME
from .ui import UserInterface, ConsoleUI
from .state_manager import StateManager
from .reporting import generate_timeline_report
from .sandbox import SandboxManager

# Configuration for the workflow steps
@dataclass
class WorkflowStep:
    name: str
    role_description: str
    input_keys: List[str] 
    output_keys: List[str] 
    arg_name: str
    model_arg_name: str
    fallback_tool_arg_name: str
    fallback_model_arg_name: str
    tool: Optional[str]
    model: Optional[str]
    fallback_tool: Optional[str]
    fallback_model: Optional[str]
    max_retries: int  
    transitions: List[Dict[str, str]]
    timeout: Optional[int]
    user_review_required: bool # New field
    read_only: bool # New field

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'WorkflowStep':
        model_arg_name = data["arg_name"].replace("-tool", "-model")
        fallback_tool_arg_name = data["arg_name"].replace("-tool", "-fallback-tool")
        fallback_model_arg_name = data["arg_name"].replace("-tool", "-fallback-model")
        
        # Backward compatibility for input_key/output_key
        input_keys = data.get("input_keys", [])
        if not input_keys and "input_key" in data:
            input_keys.append(data["input_key"])
        if "context_refs" in data: # Merge context_refs into input_keys
             for ref in data["context_refs"]:
                 if ref not in input_keys:
                     input_keys.append(ref)

        output_keys = data.get("output_keys", [])
        if not output_keys and "output_key" in data:
            output_keys.append(data["output_key"])

        return cls(
            name=data["name"],
            role_description=data["role_description"],
            input_keys=input_keys,
            output_keys=output_keys,
            arg_name=data["arg_name"],
            model_arg_name=model_arg_name,
            fallback_tool_arg_name=fallback_tool_arg_name,
            fallback_model_arg_name=fallback_model_arg_name,
            tool=data.get("tool"),
            model=data.get("model"),
            fallback_tool=data.get("fallback_tool"),
            fallback_model=data.get("fallback_model"),
            max_retries=data.get("max_retries", 0),
            transitions=data.get("transitions", []),
            timeout=data.get("timeout"),
            user_review_required=data.get("user_review_required", False),
            read_only=data.get("read_only", False)
        )

# Load full set for argparse definition only
_full_default_config_path = os.path.join(os.path.dirname(__file__), DEFAULT_CONFIG_FILENAME)
with open(_full_default_config_path, 'r') as f:
    _full_default_config = json.load(f)
ALL_POSSIBLE_STEPS = [WorkflowStep.from_dict(step_data) for step_data in _full_default_config.get("steps", [])]

TOOL_FACTORIES = {
    "gemini": create_gemini_agent,
    "amazonq": create_amazonq_agent,
    "opencode": create_opencode_agent,
    "aider": create_aider_agent,
    "github-copilot": create_github_copilot_agent,
    "gh-copilot": create_github_copilot_agent
}

class SoftwareCompanyWorkflow:
    def __init__(self, task_id: str, 
                 steps: List[WorkflowStep],
                 state_directory: str,
                 role_tool_map: Dict[str, str], 
                 role_model_map: Dict[str, str], 
                 role_fallback_tool_map: Dict[str, str],
                 role_fallback_model_map: Dict[str, str],
                 default_tool: str,
                 work_dir: Optional[str] = None,
                 max_workflow_iterations: int = 20,
                 default_timeout: Optional[int] = None,
                 ui: Optional[UserInterface] = None):
        self.task_id = task_id
        self.steps = steps
        self.state_directory = state_directory
        self.role_tool_map = role_tool_map
        self.role_model_map = role_model_map
        self.role_fallback_tool_map = role_fallback_tool_map
        self.role_fallback_model_map = role_fallback_model_map
        self.default_tool = default_tool
        self.work_dir = work_dir
        self.max_workflow_iterations = max_workflow_iterations
        self.default_timeout = default_timeout
        
        if '-' in self.task_id:
            raise ValueError(f"Task ID '{self.task_id}' cannot contain hyphens (-).")
        
        self.ui = ui or ConsoleUI()
        self.state_manager = StateManager(self.task_id, self.state_directory)
        self.state = self.state_manager.state
        self.task_dir = self.state_manager.task_dir

        # Initialize Sandbox Manager
        # We load the config again or we can assume it's passed somewhere?
        # Actually config is not passed to Workflow init. We can load it.
        # Ideally Workflow should accept config dict, but for now we load it.
        # Optimization: Pass config to __init__ if possible, but let's stick to existing pattern or load it.
        # Since we don't have the config object here, we'll load it.
        self.config = load_config() # This might be redundant if we already have it in main.py, but safe.
        self.sandbox_manager = SandboxManager(self.config, self.task_id, self.work_dir or os.getcwd())

        idx = self.state_manager.get_current_step_index()
        # Check if we have progress (index > 0 or existing context like user_request)
        if idx > 0 or self.state_manager.get_context("user_request"):
            if idx < len(self.steps):
                self.ui.print_info(f"--- Resuming task '{self.task_id}' from step {idx} ({self.steps[idx].name}) ---")
            else:
                self.ui.print_info(f"--- Task '{self.task_id}' previously completed ---")

    def _resolve_input(self, input_key: str) -> Optional[str]:
        """Resolves input data, handling file references. Returns None if key not found or explicitly null."""
        data = self.state_manager.get_context(input_key)
        if data is None: # Explicit check for None
            return None
        
        if isinstance(data, dict) and data.get("type") == "file":
            self.ui.print_input_loading(input_key, data['path'])
            return self.state_manager.read_context_from_file(data["path"])
        
        return str(data)

    async def _execute_step(self, step: WorkflowStep, agent_name: str, tool_name: str, model_name: str) -> Dict[str, str]:
        if tool_name not in TOOL_FACTORIES:
            raise ValueError(f"Unknown tool '{tool_name}'")

        factory = TOOL_FACTORIES[tool_name]
        self.ui.print_assignment(step.name, tool_name, model_name)

        # Determine timeout: Step specific > Global Default
        timeout_val = step.timeout if step.timeout is not None else self.default_timeout
        if timeout_val:
            self.ui.print_timeout(timeout_val)

        agent = factory(
            name=agent_name,
            model=model_name,
            work_dir=self.work_dir,
            timeout=timeout_val,
            read_only=step.read_only,
            sandbox_manager=self.sandbox_manager
        )
        
        # Only enable streaming for TUI, disable for Console/Non-TUI
        if not isinstance(self.ui, ConsoleUI):
            agent.set_stream_callback(lambda line, source: self.ui.print_stream_line(line, source))
        
        # Resolve Inputs
        inputs_content = ""
        for key in step.input_keys:
            val = self._resolve_input(key)
            if val is not None:
                inputs_content += f"\n\nINPUT ({key.upper().replace('_', ' ')}):\n{val}"
            # If val is None, we just skip it (it's null/missing context)

        # Construct Output Requirement instruction
        output_req = ", ".join(step.output_keys)
        json_format_hint = "You MUST return the output as a valid JSON object. The keys should be: " + output_req

        read_only_instr = ""
        if step.read_only:
             read_only_instr = "You are in READ-ONLY mode. DO NOT modify any files. DO NOT create new files. Only analyze the provided context and return the requested JSON output.\n"

        prompt = (
            f"{read_only_instr}"
            f"ROLE: {step.role_description}\n"
            f"{inputs_content}\n\n"
            f"INSTRUCTIONS: Please produce your output.\n"
            f"{json_format_hint}\n"
            f"Ensure the JSON is well-formed."
        )
        
        msg = TextMessage(content=prompt, source="workflow_engine")

        self.ui.print_tool_start(step.name, tool_name)
        try:
            result = await agent.run(task=msg)
            # Ensure we end with a newline after streaming output
            self.ui.print_newline()
        finally:
            pass
        
        if result.messages:
            raw_output = result.messages[-1].content
            # Parse JSON
            try:
                # Try to find JSON block if mixed with text
                start = raw_output.find('{')
                end = raw_output.rfind('}') + 1
                if start != -1 and end != -1:
                    json_str = raw_output[start:end]
                    parsed = json.loads(json_str)
                    return parsed
                else:
                    # Fallback: if single output key expected, treat whole text as value
                    if len(step.output_keys) == 1:
                        return {step.output_keys[0]: raw_output}
                    else:
                        raise ValueError("Could not find JSON object in response")
            except json.JSONDecodeError:
                 if len(step.output_keys) == 1:
                     return {step.output_keys[0]: raw_output}
                 raise ValueError(f"Failed to parse JSON output: {raw_output[:100]}...")
        else:
            raise RuntimeError(f"Agent {tool_name} produced no messages.")

    def revert_last_step(self):
        """Reverts the last completed step manually."""
        idx = self.state_manager.get_current_step_index()
        if idx == 0:
            self.ui.print_warning("No steps completed to revert.")
            return

        # The step to revert is at index idx - 1
        step_to_revert_idx = idx - 1
        step = self.steps[step_to_revert_idx]
        
        self.ui.print_info(f"Reverting Step {step_to_revert_idx + 1}: {step.name}")
        
        # Remove all outputs generated by this step
        for key in step.output_keys:
            output_entry = self.state_manager.get_context(key)
            if isinstance(output_entry, dict) and output_entry.get("type") == "file":
                filepath = output_entry.get("path")
                try:
                    if self.state_manager.delete_file(filepath):
                        self.ui.print_info(f"Deleted output file: {filepath}")
                except Exception as e:
                    self.ui.print_warning(f"Failed to delete {filepath}: {e}")
            
            self.state_manager.remove_context(key)
        
        # Decrement index
        self.state_manager.set_current_step_index(step_to_revert_idx)
        self.state_manager.save_state()
        self.ui.print_success(f"Step {step.name} reverted. You can now resume the task to retry this step.")

    def _determine_next_step(self, current_step: WorkflowStep, output_dict: Dict[str, str], current_index: int) -> int:
        """Determines the next step index based on transitions and output dictionary."""
        if not current_step.transitions:
            return current_index + 1
        
        self.ui.print_info(f"Checking transitions for step '{current_step.name}'...")
        for transition in current_step.transitions:
            # New field: condition_key. If not present, default to the first output key (backward compat logic)
            condition_key = transition.get("condition_key", current_step.output_keys[0] if current_step.output_keys else None)
            
            if not condition_key or condition_key not in output_dict:
                continue

            output_val = output_dict[condition_key]
            if not isinstance(output_val, str):
                output_val = str(output_val)

            condition_type = transition.get("condition_type", "contains")
            condition_value = transition.get("condition_value", "").lower()
            target_step_name = transition.get("target_step")
            
            match = False
            if condition_type == "contains":
                if condition_value in output_val.lower():
                    match = True
            
            if match:
                self.ui.print_info(f"Transition match found: Key '{condition_key}' contains '{condition_value}'. Jumping to '{target_step_name}'.")
                # Find index of target step
                for i, s in enumerate(self.steps):
                    if s.name == target_step_name:
                        return i
                self.ui.print_warning(f"Target step '{target_step_name}' not found. Continuing to next step.")
        
        return current_index + 1

    def jump_to_step(self, step_name: str):
        """Jumps to a specific step by name, verifying context prerequisites."""
        target_index = -1
        target_step = None
        
        # 1. Find the step
        for i, step in enumerate(self.steps):
            if step.name.lower() == step_name.lower():
                target_index = i
                target_step = step
                break
        
        if target_index == -1:
            available_steps = ", ".join([s.name for s in self.steps])
            raise ValueError(f"Step '{step_name}' not found. Available steps: {available_steps}")

        self.ui.print_info(f"Attempting jump to step '{target_step.name}' (Index {target_index})...")

        # 2. Check for inputs
        # Special case: First step usually takes user_request which is injected by run() if present.
        # If jumping to first step, we assume it's fine or run() will catch missing user_request.
        if target_index > 0:
            missing_inputs = []
            for key in target_step.input_keys:
                if self.state_manager.get_context(key) is None:
                    missing_inputs.append(key)
            
            if missing_inputs:
                raise ValueError(f"Cannot jump to '{target_step.name}'. Missing required context inputs: {', '.join(missing_inputs)}. Please ensure previous steps have been completed or context is manually populated.")

        # 3. Update state
        self.state_manager.set_current_step_index(target_index)
        self.state_manager.save_state()
        self.ui.print_success(f"Successfully jumped to step '{target_step.name}'. Ready to run.")

    async def run(self, user_request: str = "", single_step: bool = False):
        # Start Sandbox if enabled
        if self.sandbox_manager.is_active():
            self.ui.print_info(f"Starting sandbox environment...")
            self.sandbox_manager.start()
        
        try:
            if self.state_manager.get_context("user_request") is None and user_request:
                self.state_manager.set_context("user_request", user_request)
                self.state_manager.save_state()
            elif self.state_manager.get_context("user_request") is None:
                 self.ui.print_error("No user request provided for new task.")
                 return

            iteration_count = 0
            
            while self.state_manager.get_current_step_index() < len(self.steps):
                if iteration_count >= self.max_workflow_iterations:
                     self.ui.print_warning(f"Max workflow iterations ({self.max_workflow_iterations}) reached. Halting loop to prevent infinite cycles.")
                     generate_timeline_report(self.task_id, self.state_manager)
                     break
                iteration_count += 1

                i = self.state_manager.get_current_step_index()
                step = self.steps[i]

                tool_name = self.role_tool_map.get(step.name, self.default_tool).lower()
                model_name = self.role_model_map.get(step.name, None)

                self.ui.print_step_header(i+1, step.name, self.work_dir)

                agent_name = f"{step.name.lower().replace(' ', '_')}_{self.task_id}"

                output_dict = {}
                success = False

                start_time = datetime.now()
                used_tool = tool_name

                # Retry Loop
                max_attempts = step.max_retries + 1
                for attempt in range(max_attempts):
                    try:
                        output_dict = await self._execute_step(step, agent_name, tool_name, model_name)
                        success = True
                        break
                    except Exception as e:
                        self.ui.print_error(f"Error executing step (Attempt {attempt+1}/{max_attempts}): {e}")

                        if attempt == max_attempts - 1:
                            # Fallback logic
                            fallback_tool = self.role_fallback_tool_map.get(step.name)
                            fallback_model = self.role_fallback_model_map.get(step.name)
                            if fallback_tool:
                                self.ui.print_warning(f"--- ATTEMPTING FALLBACK: {fallback_tool} (Model: {fallback_model or 'Default'}) ---")
                                try:
                                    used_tool = fallback_tool
                                    agent_name_fb = f"{agent_name}_fallback"
                                    output_dict = await self._execute_step(step, agent_name_fb, fallback_tool, fallback_model)
                                    success = True
                                except Exception as e_fb:
                                    self.ui.print_error(f"Fallback also failed: {e_fb}. Halting.")
                                    end_time = datetime.now()
                                    duration_sec = (end_time - start_time).total_seconds()
                                    self.state_manager.add_timeline_entry({
                                        "sequence_id": len(self.state_manager.get_timeline()) + 1,
                                        "step_name": step.name,
                                        "role_tool": used_tool,
                                        "status": "FAILED",
                                        "start_time": start_time.isoformat(timespec='seconds'),
                                        "end_time": end_time.isoformat(timespec='seconds'),
                                        "duration": f"{int(duration_sec // 60):02d}:{int(duration_sec % 60):02d}"
                                    })
                                    self.state_manager.save_state()
                                    generate_timeline_report(self.task_id, self.state_manager)
                                    return
                            else:
                                self.ui.print_error("No fallback configured and retries exhausted. Halting.")
                                end_time = datetime.now()
                                duration_sec = (end_time - start_time).total_seconds()
                                self.state_manager.add_timeline_entry({
                                    "sequence_id": len(self.state_manager.get_timeline()) + 1,
                                    "step_name": step.name,
                                    "role_tool": used_tool,
                                    "status": "FAILED",
                                    "start_time": start_time.isoformat(timespec='seconds'),
                                    "end_time": end_time.isoformat(timespec='seconds'),
                                    "duration": f"{int(duration_sec // 60):02d}:{int(duration_sec % 60):02d}"
                                })
                                self.state_manager.save_state()
                                generate_timeline_report(self.task_id, self.state_manager)
                                return

                end_time = datetime.now()
                duration_sec = (end_time - start_time).total_seconds()

                self.state_manager.add_timeline_entry({
                    "sequence_id": len(self.state_manager.get_timeline()) + 1,
                    "step_name": step.name,
                    "role_tool": used_tool,
                    "status": "COMPLETED",
                    "start_time": start_time.isoformat(timespec='seconds'),
                    "end_time": end_time.isoformat(timespec='seconds'),
                    "duration": f"{int(duration_sec // 60):02d}:{int(duration_sec % 60):02d}"
                })

                if not success:
                    return

                self.ui.print_output(step.name, output_dict)

                # WRITE OUTPUTS TO FILES
                written_files = []
                for key, val in output_dict.items():
                    if isinstance(val, (dict, list)):
                        content = json.dumps(val, indent=2)
                    else:
                        content = str(val)

                    file_path = self.state_manager.write_context_to_file(i + 1, step.name, key, content)
                    written_files.append((key, file_path))
                    
                    # Update state with file reference
                    self.state_manager.set_context(key, {
                        "type": "file",
                        "path": file_path
                    })
                
                # Notify UI of written artifacts
                self.ui.print_artifacts(step.name, written_files)
                
                # Display summary of completed step outputs
                self.ui.print_step_outputs_summary(step.name, written_files)
                
                # Determine next step
                next_index = self._determine_next_step(step, output_dict, i)

                self.state_manager.set_current_step_index(next_index)
                self.state_manager.save_state()

                # USER INTERVENTION
                if step.user_review_required:
                    # Iterate through only the output keys of the current step
                    current_step_output_paths = []
                    for key in step.output_keys:
                        output_entry = self.state_manager.get_context(key)
                        if isinstance(output_entry, dict) and output_entry.get("type") == "file":
                            current_step_output_paths.append((key, output_entry["path"]))

                    self.ui.print_user_intervention(step.name, current_step_output_paths)
                    self.ui.input("Press Enter to continue...")
                    self.ui.print_info("Resuming workflow...")

                if single_step:
                    generate_timeline_report(self.task_id, self.state_manager)
                    self.ui.print_info("--- Stopping after single step execution as requested ---")
                    return

            if self.state_manager.get_current_step_index() >= len(self.steps):
                generate_timeline_report(self.task_id, self.state_manager)
                self.ui.print_success(f"=== WORKFLOW FOR TASK '{self.task_id}' COMPLETED SUCCESSFULLY ===")
        finally:
            if self.sandbox_manager.is_active():
                 self.ui.print_info(f"Stopping sandbox environment...")
                 self.sandbox_manager.stop()



from typing import Protocol, List, Optional
from rich.console import Console
from rich.panel import Panel
from rich.markdown import Markdown
from rich.syntax import Syntax
from rich.theme import Theme
from rich.json import JSON
from rich.table import Table
import json

class UserInterface(Protocol):
    def print_step_header(self, step_number: int, step_name: str, work_dir: str = None):
        ...

    def print_assignment(self, role: str, tool: str, model: str):
        ...

    def print_timeout(self, timeout_val: int):
        ...

    def print_input_loading(self, input_key: str, path: str):
        ...

    def print_tool_start(self, step_name: str, tool_name: str):
        ...

    def start_spinner(self, message: str):
        ...

    def stop_spinner(self):
        ...

    def print_output(self, step_name: str, output: dict):
        ...

    def print_error(self, message: str):
        ...

    def print_warning(self, message: str):
        ...

    def print_info(self, message: str):
        ...

    def print_success(self, message: str):
        ...

    def print_stream_line(self, line: str, source: str = "stdout"):
        ...

    def input(self, prompt: str) -> str:
        ...

    def print_user_intervention(self, step_name: str, files: list):
        ...

    def print_artifacts(self, step_name: str, artifacts: list):
        ...

    def print_step_outputs_summary(self, step_name: str, artifacts: list):
        ...

    def print_newline(self):
        ...


class ConsoleUI:
    def __init__(self):
        self.console = Console(theme=Theme({
            "info": "dim cyan",
            "warning": "magenta",
            "error": "bold red",
            "success": "bold green",
            "step": "bold blue",
            "json.key": "bold cyan",
            "json.str": "green",
            "highlight": "bold yellow",
            "role": "bold magenta",
            "tool": "bold cyan",
            "file": "underline blue",
        }))
        self.status = None

    def print_artifacts(self, step_name: str, artifacts: list):
        self.console.print(f"[bold]Generated Artifacts:[/bold]")
        for key, path in artifacts:
            self.console.print(f" - [bold cyan]{key}[/bold cyan]: [file]{path}[/file]")
        self.console.print()

    def print_step_outputs_summary(self, step_name: str, artifacts: list):
        """Display output file paths from a completed step in a summary section."""
        self.console.print(f"[bold green]✓ Completed:[/bold green] {step_name}")
        for key, path in artifacts:
            self.console.print(f"  [dim]→[/dim] [bold cyan]{key}[/bold cyan]: [file]{path}[/file]")
        self.console.print()

    def print_newline(self):
        self.console.print()

    def print_step_header(self, step_number: int, step_name: str, work_dir: str = None):
        title = f"[step]STEP {step_number}: {step_name}[/step]"
        content = ""
        if work_dir:
            content = f"Working Directory: [dim]{work_dir}[/dim]"
        else:
            content = "Running..."

        self.console.print()
        self.console.print(Panel(content, title=title, border_style="blue", expand=False))

    def print_assignment(self, role: str, tool: str, model: str):
        grid = Table.grid(expand=True)
        grid.add_column(justify="right", style="dim")
        grid.add_column(justify="left")

        grid.add_row("Role:", f"[role]{role}[/role]")
        grid.add_row("Tool:", f"[tool]{tool}[/tool]")
        grid.add_row("Model:", f"[dim]{model or 'Default'}[/dim]")

        self.console.print(Panel(grid, title="[bold]Task Assignment[/bold]", border_style="cyan", expand=False))

    def print_timeout(self, timeout_val: int):
        self.console.print(f"   [highlight]⏳ Timeout:[/highlight] [bold]{timeout_val}s[/bold]")

    def print_input_loading(self, input_key: str, path: str):
        self.console.print(f"   [dim]📥 Loading input[/dim] [bold]{input_key}[/bold] [dim]from[/dim] [file]{path}[/file]")

    def print_tool_start(self, step_name: str, tool_name: str):
        self.console.print(f"[bold green]🚀 Starting execution:[/bold green] [bold]{step_name}[/bold] [dim](via {tool_name})[/dim]...")
        self.console.print("[dim]   (Streaming output below)[/dim]")

    def start_spinner(self, message: str):
        self.status = self.console.status(message)
        self.status.start()

    def stop_spinner(self):
        if self.status:
            self.status.stop()
            self.status = None

    def print_output(self, step_name: str, output: dict):
        self.console.print(f"\n[success]--- Output from {step_name} ---[/success]")

        for key, val in output.items():
            self.console.print(f"[bold cyan]>> {key}[/bold cyan]")
            if isinstance(val, str):
                # Check if it is a JSON string
                stripped = val.strip()
                if (stripped.startswith("{") and stripped.endswith("}")) or \
                   (stripped.startswith("[") and stripped.endswith("]")):
                    try:
                        self.console.print_json(val)
                        self.console.print()
                        continue
                    except json.JSONDecodeError:
                        pass

                # Check if it looks like code (simple heuristic)
                if stripped.startswith(("def ", "class ", "import ", "from ", "<!DOCTYPE", "<html>")):
                     # Use 'python' as default guess, but rich can guess too if we pass no lexer?
                     # Syntax(val, lexer="python")
                     # Let's try to detect based on content or just use Markdown code blocks
                     self.console.print(Markdown(f"```\n{val}\n```"))
                else:
                    self.console.print(Markdown(val))
            elif isinstance(val, (dict, list)):
                 self.console.print_json(data=val)
            else:
                 self.console.print(str(val))
            self.console.print()

    def print_error(self, message: str):
        self.console.print(f"[error]Error:[/error] {message}")

    def print_warning(self, message: str):
        self.console.print(f"[warning]Warning:[/warning] {message}")

    def print_info(self, message: str):
        self.console.print(f"[info]{message}[/info]")

    def print_success(self, message: str):
        self.console.print(f"[success]{message}[/success]")

    def print_stream_line(self, line: str, source: str = "stdout"):
        style = "dim" if source == "stdout" else "red"
        # We use markup=False to prevent treating brackets as style tags.
        # This allows raw output (including ANSI codes if supported by the terminal) to be printed safely
        # without Rich trying to interpret it as Rich markup.
        self.console.print(line, style=style, markup=False, end="")

    def input(self, prompt: str):
        return self.console.input(f"[bold yellow]{prompt}[/bold yellow]")

    def print_user_intervention(self, step_name: str, files: list):
        self.console.print(Panel(
            f"Step '[bold]{step_name}[/bold]' is paused for user review.\n" +
            "Generated output files:\n" +
            "\n".join([f" - [cyan]{key}[/cyan]: {path}" for key, path in files]) +
            "\n\nYou may edit these files now. When you are ready to proceed, press Enter...",
            title="[bold yellow]USER INTERVENTION[/bold yellow]",
            border_style="yellow"
        ))

# For backward compatibility if needed, though we will refactor usage
UI = ConsoleUI



import subprocess
import shutil
import os
import threading
import queue
from typing import Optional, List, Callable
from abc import ABC, abstractmethod

class BaseCLIWrapper(ABC):
    """
    Abstract base class for interactive CLI agent wrappers.
    Handles process execution and lifecycle.
    """

    def __init__(self, command: str, auto_approve: bool = True, model: Optional[str] = None, work_dir: Optional[str] = None, read_only: bool = False, sandbox_manager=None):
        """
        Args:
            command (str): The base command to run.
            auto_approve (bool): Whether to auto-approve tool calls (if supported).
            model (str, optional): The model ID/name to use.
            work_dir (str, optional): The working directory for the CLI process.
            read_only (bool): Whether to run in read-only mode (default: False).
            sandbox_manager (SandboxManager, optional): Manager for Docker sandbox execution.
        """
        self.command = command
        self.auto_approve = auto_approve
        self.model = model
        self.work_dir = work_dir
        self.read_only = read_only
        self.sandbox_manager = sandbox_manager
        self.session_id: Optional[str] = None
        
        # Verify command exists (only if not sandboxed)
        if not (self.sandbox_manager and self.sandbox_manager.is_active()):
            executable = self.command.split()[0]
            if not shutil.which(executable):
                raise FileNotFoundError(f"Command '{executable}' not found in PATH.")

            # Verify work_dir if provided
            if self.work_dir and not os.path.isdir(self.work_dir):
                raise NotADirectoryError(f"Working directory '{self.work_dir}' does not exist.")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def ask(self, prompt_text: str, timeout: int = 120, stream_callback: Optional[Callable[[str, str], None]] = None) -> str:
        """
        Sends a message to the CLI and returns the processed response.

        Args:
            prompt_text (str): The input to send to the CLI.
            timeout (int): Timeout in seconds.
            stream_callback (Callable): Optional callback(text, source) for real-time output.
        """
        cmd_args = self._construct_command(prompt_text)
        cwd = self.work_dir

        if self.sandbox_manager and self.sandbox_manager.is_active():
            cmd_args = self.sandbox_manager.wrap_command(cmd_args, work_dir=self.work_dir)
            cwd = None # docker exec handles working directory

        if os.getenv("DEBUG_CLI_WRAPPER"):
            print(f"[DEBUG_CLI] Executing: {' '.join(cmd_args)}")
            print(f"[DEBUG_CLI] CWD: {cwd or os.getcwd()}")

        try:
            process = subprocess.Popen(
                cmd_args,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                encoding='utf-8',
                cwd=cwd, # Execute in the specified directory
                bufsize=1 # Line buffered
            )
        except Exception as e:
            raise RuntimeError(f"Failed to execute {self.command}: {e}")

        # Queues to store output for final aggregation
        stdout_queue = queue.Queue()
        stderr_queue = queue.Queue()

        full_stdout = []
        full_stderr = []

        def reader(pipe, q, source):
            try:
                for line in pipe:
                    q.put(line)
                    if stream_callback:
                        self._process_stream_line(line, source, stream_callback)
            except ValueError:
                # Handle case where pipe is closed/invalid
                pass
            finally:
                pipe.close()

        t_stdout = threading.Thread(target=reader, args=(process.stdout, stdout_queue, "stdout"))
        t_stderr = threading.Thread(target=reader, args=(process.stderr, stderr_queue, "stderr"))

        t_stdout.start()
        t_stderr.start()

        try:
            process.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            t_stdout.join()
            t_stderr.join()

            # Collect whatever we got so far
            while not stdout_queue.empty(): full_stdout.append(stdout_queue.get())
            while not stderr_queue.empty(): full_stderr.append(stderr_queue.get())

            stdout_str = "".join(full_stdout)
            stderr_str = "".join(full_stderr)

            print(f"[WARN] Command timed out after {timeout}s")
            raise RuntimeError(f"{self.command} timed out after {timeout}s\nStdout: {stdout_str}\nStderr: {stderr_str}")

        t_stdout.join()
        t_stderr.join()

        while not stdout_queue.empty(): full_stdout.append(stdout_queue.get())
        while not stderr_queue.empty(): full_stderr.append(stderr_queue.get())

        stdout_str = "".join(full_stdout)
        stderr_str = "".join(full_stderr)

        if process.returncode != 0:
            err_msg = stderr_str if stderr_str else "(no stderr)"
            # Some CLIs use non-zero exit codes for "soft" errors which we might want to parse.
            # But generally it's an error.
            raise RuntimeError(f"{self.command} exited with code {process.returncode}:\n{err_msg}\nStdout: {stdout_str}")

        return self._parse_output(stdout_str)

    def _process_stream_line(self, line: str, source: str, callback: Callable[[str, str], None]) -> None:
        """
        Process a single line of streamed output.
        Subclasses can override this to handle JSON streams or other formats.
        """
        callback(line, source)

    @abstractmethod
    def _construct_command(self, prompt_text: str) -> List[str]:
        """
        Builds the full command list [cmd, arg1, arg2...]
        """
        pass

    @abstractmethod
    def _parse_output(self, stdout: str) -> str:
        """
        Parses the raw stdout string into a clean response.
        Should also handle session ID extraction if applicable.
        """
        pass

    def close(self) -> None:
        """Resets session state."""
        self.session_id = None



import json
import os
from typing import Dict, Any

DEFAULT_CONFIG_FILENAME = "default_config.json"
USER_CONFIG_FILENAME = "config.json"

def load_config(config_path: str = None) -> Dict[str, Any]:
    """
    Loads configuration.
    1. If config_path is provided, load it.
    2. Else, try loading 'config.json' from current working directory.
    3. Fallback to 'default_config.json' in the package directory.
    """
    # 1. Explicit path provided via CLI
    if config_path:
        if os.path.exists(config_path):
            try:
                with open(config_path, 'r') as f:
                    print(f"Loading configuration from {config_path}...")
                    return json.load(f)
            except Exception as e:
                print(f"Error loading {config_path}: {e}. Aborting.")
                raise # Explicit path failure should probably raise error
        else:
            print(f"Error: Config file '{config_path}' not found.")
            raise FileNotFoundError(f"Config file '{config_path}' not found.")

    # 2. Try loading user config from CWD
    if os.path.exists(USER_CONFIG_FILENAME):
        try:
            with open(USER_CONFIG_FILENAME, 'r') as f:
                print(f"Loading configuration from {USER_CONFIG_FILENAME} (CWD)...")
                return json.load(f)
        except Exception as e:
            print(f"Error loading {USER_CONFIG_FILENAME} from CWD: {e}. Falling back to default.")

    # 3. Load default config from package
    package_dir = os.path.dirname(__file__)
    default_config_path = os.path.join(package_dir, DEFAULT_CONFIG_FILENAME)
    print(f"Loading default config from {default_config_path}...")
    
    if os.path.exists(default_config_path):
        with open(default_config_path, 'r') as f:
            return json.load(f)
    
    # 4. Fallback hardcoded (should not happen if installed correctly)
    return {
        "state_directory": "workflow_states",
        "default_tool": "gemini",
        "sandbox": {
            "enabled": False,
            "image": "python:3.11-slim",
            "mount_path": "/workspace"
        },
        "steps": []
    }
